import treams.special
import numpy as np
from acoutreams._materialacoustics import AcousticMaterial

def mie_acoustics(l, x, *materials):
    
    # Returns acoustical Mie coefficients for a sphere

    # l - order
    # x = k0 * radius - size parameter (k0 is in air)
    # *materials - list of tuples of material parameters

    mat_sphere, mat_env = zip(*materials) 
    x_env = x * AcousticMaterial().c / mat_env[1]
    j = treams.special.spherical_jn(l, x_env, derivative=False)
    j_d = treams.special.spherical_jn(l, x_env, derivative=True)
    h = treams.special.spherical_hankel1(l, x_env)
    h_d = treams.special.spherical_hankel1_d(l, x_env)
    if np.abs(mat_sphere[2]) == 0 and np.abs(mat_sphere[1]) == 0 and np.abs(mat_sphere[0]) == np.inf:
        res = -j_d / h_d
    elif np.abs(mat_sphere[2]) == 0 and np.abs(mat_sphere[1]) == 0 and np.abs(mat_sphere[0]) == 0:
        res = -j / h
    elif np.abs(mat_sphere[1]) > 0:
        x_sphere = x * AcousticMaterial().c / mat_sphere[1]
        j1 = treams.special.spherical_jn(l, x_sphere, derivative=False)
        j1_d = treams.special.spherical_jn(l, x_sphere, derivative=True)
        if np.abs(mat_sphere[2]) == 0:
            delta = x_sphere * mat_env[0] / (x_env * mat_sphere[0]) 
            res = (delta * j1_d * j - j1 * j_d) / (j1 * h_d - delta * j1_d * h) 
        elif np.abs(mat_sphere[2]) > 0:
            x_sphere_t = x * AcousticMaterial().c / mat_sphere[2]
            j1t = treams.special.spherical_jn(l, x_sphere_t, derivative=False)
            j1t_d = treams.special.spherical_jn(l, x_sphere_t, derivative=True)
            d44 = (l * (l + 1) - 0.5 * x_sphere_t**2) * j1 - 2 * x_sphere * j1_d
            d33 = (l * (l + 1) - 0.5 * x_sphere_t**2 - 1) * j1t - x_sphere_t * j1t_d
            d34 = x_sphere * j1_d - j1
            d43 = l * (l + 1) * (x_sphere_t * j1t_d - j1t)
            d24 = x_sphere * j1_d
            d23 = l * (l + 1) * j1t
            zeta = (d44*d33 - d34*d43) / (d24*d33 - d34*d23)
            res = (-(mat_env[0]/mat_sphere[0] * 0.5 * x_sphere_t**2/x_env * j + zeta * j_d) / 
                (mat_env[0]/mat_sphere[0] * 0.5 * x_sphere_t**2/x_env * h + zeta * h_d))

    return res * np.ones(2*l+1)

def mie_acoustics_cyl(kz, m, k0, radii, *materials):
    
    # Returns acoustical Mie coefficients for an infinite cylinder (along z-axis)

    # kz - z-component of the wave vector
    # m - angular momentum
    # k0 - wave number in air
    # radii - radius of a cylinder
    # *materials - list of tuples of material parameters

    mat_cyl, mat_env = zip(*materials) 
    k_env = k0 * AcousticMaterial().c / mat_env[1]
    krho_cyl = np.sqrt(k_cyl * k_cyl - kz * kz)
    krho_env = np.sqrt(k_env * k_env - kz * kz)
    x_cyl = krho_cyl * radii
    x_env = krho_env * radii
    j = treams.special.jv(m, x_env)
    j_d = treams.special.jv_d(m, x_env)
    h = treams.special.hankel1(m, x_env)
    h_d = treams.special.hankel1_d(m, x_env)
    
    if np.abs(mat_cyl[2]) == 0 and np.abs(mat_cyl[1]) == 0 and np.abs(mat_cyl[0]) == np.inf:
        res = -j_d / h_d
    elif np.abs(mat_cyl[2]) == 0 and np.abs(mat_cyl[1]) == 0 and np.abs(mat_cyl[0]) == 0:
        res = -j / h
    if np.abs(mat_cyl[1]) > 0:
        k_cyl = k0 * AcousticMaterial().c / mat_cyl[1]
        j1 = treams.special.jv(m, x_cyl)
        j1_d = treams.special.jv_d(m, x_cyl)
        if np.abs(mat_cyl[2]) == 0:
            delta = x_cyl * mat_env[0] / (x_env * mat_cyl[0]) 
            res = (delta * j1_d * j - j1 * j_d) / (j1 * h_d - delta * j1_d * h)
        elif np.abs(mat_cyl[2]) > 0:
            k_cyl_t = k0 * AcousticMaterial().c / mat_cyl[2]
            krho_cyl_t = np.sqrt(k_cyl_t * k_cyl_t - kz * kz)
            x_cyl_t = krho_cyl_t * radii
            j1t = treams.special.jv(m, x_cyl_t)
            j1t_d = treams.special.jv_d(m, x_cyl_t)
            j1t_dd = (-j1t_d + (x_cyl_t**2 + m**2) * j1t) / x_cyl_t
            j1_dd = (-j1_d + (x_cyl**2 + m**2) * j1) / x_cyl
            matrix = np.zeros((3, 3), complex)
            rhs = np.zeros(3, complex)
            lam_0 = mat_env[0] * mat_env[1]**2
            mu_1 = mat_cyl[0] * mat_cyl[2]**2
            lam_1 = mat_cyl[0] * mat_cyl[1]**2 - 2 * mu_1
            matrix[0][1] = 2j * kz * krho_cyl/k_cyl * j1_d
            matrix[0][2] = (krho_cyl_t**2 - kz**2)/k_cyl_t * j1t_d
            matrix[1][0] = k_env * lam_0 * h
            matrix[1][1] = 2 * mu_1 * krho_cyl**2/k_cyl * j1_dd - k_cyl * lam_1 * j1
            matrix[1][2] = 2j * mu_1 * kz * krho_cyl_t/k_cyl_t * j1t_dd
            rhs[1] = -np.power(-1j, m) * k_env * lam_0 * j
            matrix[2][0] = -krho_env/k_env * h_d
            matrix[2][1] = krho_cyl/k_cyl * j1_d
            matrix[2][2] = 1j * kz/k_cyl_t * j1t_d
            rhs[2] = np.power(-1j, m) * krho_env/k_env * j_d
            res = np.linalg.solve(matrix, rhs)
            res = res[0] * np.power(1j, m)
    return res

def fresnel_acoustics(kzs, rhos):
    res = np.zeros((2, 2, 1, 1), complex)
    res[0][0][0][0] = 2 * rhos[0] * kzs[1] / (rhos[0] * kzs[1] + rhos[1] * kzs[0])
    res[1][1][0][0] = 2 * rhos[1] * kzs[0] / (rhos[0] * kzs[1] + rhos[1] * kzs[0])
    res[0][1][0][0] = (-rhos[0] * kzs[1] + rhos[1] * kzs[0]) / (rhos[0] * kzs[1] + rhos[1] * kzs[0])
    res[1][0][0][0] = -res[0][1]
    return res


"""ACOUTREAMS: A Python package for acoustic scattering based on the T-matrix method.

.. currentmodule:: acoutreams

Classes
=======

The top-level classes and functions allow a high-level access to the functionality.

Basis sets
----------

   ScalarCylindricalWaveBasis
   ScalarPlaneWaveBasisByUnitVector
   ScalarPlaneWaveBasisByComp
   ScalarSphericalWaveBasis

Matrices and Arrays
-------------------

   AcousticPhysicsArray
   AcousticSMatrix
   AcousticSMatrices
   AcousticTMatrix
   AcousticTMatrixC

Other
-----

   Lattice
   AcousticMaterial

Functions
=========

   vfield
   pfield
   ffamplitude
   expand
   expandlattice
   permute
   plane_wave
   rotate
   translate

"""

_version__ = "0.1.0"

from scipy.special import (  # noqa: F401
    hankel1,
    hankel2,
    jv,
    spherical_jn,
    spherical_yn,
    yv,
)

from treams.special import(   # noqa: F401
    spherical_jn_d,
    spherical_yn_d,
    sph_harm,
    lpmv,
    incgamma,
    intkambe,
    wignersmalld,
    wignerd,
    wigner3j,
    pi_fun,
    tau_fun,
    car2cyl,
    car2sph,
    cyl2car,
    cyl2sph,
    sph2car,
    sph2cyl,
    vcar2cyl,
    vcar2sph,
    vcyl2car,
    vcyl2sph,
    vsph2car,
    vsph2cyl,
    car2pol,
    pol2car,
    vcar2pol,
    vpol2car,
    )

from treams.misc import(  # noqa: F401
    pickmodes,
    wave_vec_z,
    firstbrillouin1d,
    firstbrillouin2d,
    firstbrillouin3d,
)

from treams._lattice import *   # noqa: F401

from acoutreams._wavesacoustics import *  # noqa: F401

from acoutreams._materialacoustics import AcousticMaterial  # noqa: F401

from acoutreams._smatrixacoustics import (  # noqa: F401
    AcousticSMatrices,
    AcousticSMatrix,
    poynting_avg_z,
)

from acoutreams.scw import *

from acoutreams._coreacoustics import (  # noqa: F401
    ScalarCylindricalWaveBasis,
    AcousticPhysicsArray,
    ScalarPlaneWaveBasisByComp,
    ScalarPlaneWaveBasisByUnitVector,
    ScalarSphericalWaveBasis,
)

from acoutreams._tmatrixacoustics import (  # noqa: F401
    AcousticTMatrix,
    AcousticTMatrixC,
    cylindrical_wave_scalar,
    plane_wave_scalar,
    plane_wave_angle_scalar,
    spherical_wave_scalar,
)

from acoutreams.coeffsacoustics import *  # noqa: F401 

from acoutreams.spw import *  # noqa: F401

from acoutreams._operatorsacoustics import (  # noqa: F401
    PField,
    VField,
    FFAmplitude,
    Expand,
    ExpandLattice,
    Permute,
    Rotate,
    Translate,
    vfield,
    pfield,
    ffamplitude,
    expand,
    expandlattice,
    permute,
    rotate,
    translate,
)

from acoutreams.ssw import *



import treams.special as sc
import numpy as np
import cmath
import scipy.special as ss

def spw_Psi(kx, ky, kz, x, y, z):
    """Scalar plane wave"""
    return np.exp(1j * (kx * x + ky * y + kz * z))


def vpw_L(kx, ky, kz, x, y, z):
    """Longitudinal vector plane wave"""
    k = np.sqrt(kx * kx + ky * ky + kz * kz)
    phase = np.exp(1j * (kx * x + ky * y + kz * z))
    if k == 0:
        return np.asarray([np.nan, np.nan, np.nan]).T
    else:
        return 1j * np.asarray([kx * phase, ky * phase, kz * phase], np.complex128).T / k

def ssw_Psi(l, m, kr, theta, phi):
     """Singular scalar spherical wave"""
     return sc.sph_harm(m, l, phi, theta) * sc.spherical_hankel1(l, kr)

def ssw_PsiFF(l, m, kx, ky, kz, theta, phi):
     """Far-field amplitude of singular scalar spherical wave"""
     return (sc.sph_harm(m, l, phi, theta)
             * np.exp(-1j * ((kx * np.cos(phi) + ky * np.sin(phi)) * np.sin(theta) + kz * np.cos(theta)))
             * np.power(-1j, l + 1))

def ssw_rPsi(l, m, kr, theta, phi):
     """Regular scalar spherical wave"""
     return sc.sph_harm(m, l, phi, theta) * sc.spherical_jn(l, kr)


def scw_Psi(kz, m, krr, phi, z):
     """Singular scalar cylindrical wave"""
     return np.exp(1j * (m * phi + kz * z)) * sc.hankel1(m, krr)

def scw_PsiFF(kz, m, krhox, krhoy, phi, z, krho):
     """Far-field amplitude of singular scalar cylindrical wave"""
     return (np.exp(1j * m * phi) * 
             np.sqrt(2 / (np.pi * krho)) *
             np.exp(-1j * ((krhox * np.cos(phi) + krhoy * np.sin(phi)))) *
             np.power(-1j, m) * 
             np.exp(-1j * np.pi/4))

def scw_rPsi(kz, m, krr, phi, z):
     """Regular scalar cylindrical wave"""
     return np.exp(1j * (m * phi + kz * z)) * sc.jv(m, krr)


def vsw_L(l, m, kr, theta, phi):
     """Longitudinal singular vector spherical wave"""
     return np.transpose(sc.vsh_Z(l, m, theta, phi).T * sc.spherical_hankel1_d(l, kr)  + np.sqrt(
         l * (l + 1)
     ) * sc.vsh_Y(l, m, theta, phi).T * sc.spherical_hankel1(l, kr) / kr) * (-1.0j)


def vsw_rL(l, m, kr, theta, phi):
     """Longitudinal regular vector spherical wave"""
     res = []
     for i, x in enumerate(kr):
            k = i
            if x == 0:
                if l[i] != 1:
                    val = np.zeros(3, complex).T
                    res.append(val)
                else:
                   val = np.transpose(sc.vsh_Z(1, m[k], 0, 0).T  
                                        + np.sqrt(2) * sc.vsh_Y(1, m[k], 0, 0).T) * (1/3) * (-1.0j)
                   res.append(val)
            else:      
                val = np.transpose(sc.vsh_Z(l[k], m[k], theta[i], phi[i]).T * sc.spherical_jn_d(l[k], kr[i])  + np.sqrt(
                    l[k] * (l[k] + 1)
                ) * sc.vsh_Y(l[k], m[k], theta[i], phi[i]).T * sc.spherical_jn(l[k], kr[i]) / kr[i]) * (-1.0j)
                res.append(val)
     return np.array(res)


def vcw_L(kz, m, krr, phi, z, k, krho):
     """Longitudinal singular vector cylindrical wave"""
     return np.transpose(
         np.array(
             [
                 sc.hankel1_d(m, krr) * krho / k  * np.exp(1j * (m * phi + kz * z)),
                 1j * m * sc.hankel1(m, krr) / krr * krho / k  * np.exp(1j * (m * phi + kz * z)),
                 1j * kz / k * sc.hankel1(m, krr) * np.exp(1j * (m * phi + kz * z)),
             ]
         )
     )

def vcw_rL(kz, m, krr, phi, z, k, krho):
     """Longitudinal regular vector cylindrical wave"""
     res = []
     for i, x in enumerate(krr):
            if x == 0:
                if (abs(m[i]) != 1) and (m[i] != 0):
                    val = np.zeros(3, complex).T
                    res.append(val)
                elif abs(m[i]) == 1:
                   val = np.array(
                                  [
                       0.5 * krho[i] / k[i],
                       0.5 * 1j * m[i] * krho[i] / k[i],
                       0. 
                       ]).T * np.exp(1j * (m[i] * phi[i] + kz[i] * z[i]))
                   res.append(val)
                elif m[i] == 0:
                    val = np.array(
                        [
                            0., 
                            0., 
                            1j * kz[i] / k[i] * np.exp(1j * (m[i] * phi[i] + kz[i] * z[i]))
                            ]
                            ).T
                    res.append(val)
            else:      
                val = np.array(
                [
                    sc.jv_d(m[i], krr[i]) * krho[i] / k[i],
                    1j * m[i] * sc.jv(m[i], krr[i]) / krr[i] * krho[i] / k[i],
                    1j * kz[i] / k[i] * sc.jv(m[i], krr[i]),
                ] 
            ).T * np.exp(1j * (m[i] * phi[i] + kz[i] * z[i]))
                res.append(val)
     return np.array(res)


def _tl_ssw_helper(l, m, lambda_, mu, p, q):
    """Helper function for the translation coefficient of scalar and longitudinal spherical waves"""
    if (
        p < max(abs(m + mu), abs(l - lambda_))
        or p > abs(l + lambda_)
        or q < abs(l - lambda_)
        or q > abs(l + lambda_)
        or (q + l + lambda_) % 2 != 0
    ):
        return 0
    return (
        (2 * p + 1)
        * np.power(1j, lambda_ - l + p)
        * np.sqrt(ss.gamma(p - m - mu + 1) / ss.gamma(p + m + mu + 1))
        * sc.wigner3j(l, lambda_, p, m, mu, -(m + mu))
        * sc.wigner3j(l, lambda_, q, 0, 0, 0)
    )


def tl_ssw(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs):
    """Singular translation coefficient of scalar and longitudinal spherical waves"""
    pref = np.power(-1, np.abs(m)) * np.sqrt((2 * l + 1) * (2 * lambda_ + 1)) * cmath.exp(1j * (m - mu) * phi)
    res = 0.
    max_ = np.max([np.abs(int(lambda_) - int(l)), np.abs(int(m) - int(mu))])
    min_ = int(l) +  int(lambda_)
    for p in range(min_, max_ - 1, -2):
        res += (
            _tl_ssw_helper(l, m, lambda_, -mu, p, p)
            * sc.spherical_hankel1(p, kr)
            * sc.lpmv(m - mu, p, np.cos(theta), *args, **kwargs)
        )
    return res * pref


def tl_ssw_r(lambda_, mu, l, m, kr, theta, phi, *args, **kwargs):
    """Regular translation coefficient of scalar and longitudinal spherical waves"""
    pref = np.power(-1, np.abs(m)) * np.sqrt((2 * l + 1) * (2 * lambda_ + 1)) * cmath.exp(1j * (m - mu) * phi)
    res = 0.
    max_ = np.max([np.abs(int(lambda_) - int(l)), np.abs(int(m) - int(mu))])
    min_ = int(l) + int(lambda_)
    for p in range(min_, max_ - 1, -2):
        res += (
            _tl_ssw_helper(l, m, lambda_, -mu, p, p)
            * sc.spherical_jn(p, kr)
            * sc.lpmv(m - mu, p, np.cos(theta), *args, **kwargs)
        )
    return res * pref

def tl_scw(kz1, mu, kz2, m, krr, phi, z, *args, **kwargs):
    """Singular translation coefficient of scalar and longitudinal cylindrical waves"""
    if kz1 != kz2:
        return 0.+0.j
    return sc.hankel1(m - mu, krr) * np.exp(1j * ((m - mu) * phi + kz1 * z), *args, **kwargs)


def tl_scw_r(kz1, mu, kz2, m, krr, phi, z, *args, **kwargs):
    """Regular translation coefficient of scalar and longitudinal cylindrical waves"""
    if kz1 != kz2:
        return 0.+0.j
    return sc.jv(m - mu, krr) * np.exp(1j * ((m - mu) * phi + kz1 * z), *args, **kwargs)